#include "PsycoSuperRes.h"
#include <cstring>
#include <math.h>

PsycoSuperRes::PsycoSuperRes(const float * anch_features, const float *regressors, const int n_regressors, const float *global_regressor, const int step, const int border,
                             const int upscaling_factor, const float *pivots, const int n_spheres, const float *thresholds, const int *bucket_sizes, const int *acum_sizes,
                             const int patch_side, const float * bases) : Psyco (patch_side, bases)
{
    this->n_regressors = n_regressors;
    this->step = step;
    this->border = border;
    this->upscaling_factor = upscaling_factor;
    this->n_spheres = n_spheres;

    // Initializing dynamic memory allocation and copying values
    this->anchored_features = new float [this->patch_dim * this->n_regressors];
    memcpy(this->anchored_features, anch_features, patch_dim * this->n_regressors * sizeof(float));

    this->regressors = new float [this->patch_dim * this->patch_dim * this->n_regressors];
    memcpy(this->regressors, regressors, this->patch_dim * this->patch_dim * this->n_regressors * sizeof(float));

    this->global_regressor = new float [this->patch_dim * this->patch_dim];
    memcpy(this->global_regressor, global_regressor, this->patch_dim * this->patch_dim * sizeof(float));

    this ->pivots = new float [this->patch_dim * this->n_spheres];
    memcpy(this->pivots, pivots, this->patch_dim * this->n_spheres * sizeof(float) );

    this ->thresholds = new float [this->n_spheres];
    memcpy(this->thresholds, thresholds, this->n_spheres*sizeof(float));

    int n_buckets = pow (2, this->n_spheres);
    this->bucket_sizes = new int [n_buckets];
    memcpy(this->bucket_sizes, bucket_sizes, n_buckets*sizeof(int));

    // ATTENTION - DOUBLE CHECK SIZE OF THE MATRIX IN MATLAB
    this->acum_sizes = new int [n_buckets];
    memcpy(this->acum_sizes, acum_sizes, (n_buckets+1)*sizeof(int));

}

PsycoSuperRes::~PsycoSuperRes()
{
    delete[] anchored_features;
    delete[] regressors;
    delete[] global_regressor;
    delete[] pivots;
    delete[] thresholds;
    delete[] bucket_sizes;
    delete[] acum_sizes;

    //mexPrintf("It has called the destructor(PSyCoSuperRes class)");


}

void PsycoSuperRes::upscaleImagePar(float *upscaled_im, const float *image,  const int n_rows, const int n_cols)
{
    int * n_contributions;
    n_contributions = new int [n_rows*n_cols];

    float * image_acc;
    image_acc = new float [n_rows*n_cols];

#pragma omp parallel for
    for (int iterator = 0; iterator < n_rows*n_cols; ++iterator)
    {
        n_contributions[iterator] = 0;
        image_acc[iterator] = 0.0f;
    }

#pragma omp parallel for collapse(2)
    for(int gcol_it = border ; gcol_it < n_cols - border - patch_side + 1; gcol_it += step)
    {
        for ( int grow_it = border ; grow_it < n_rows - border - patch_side + 1; grow_it += step )
        {
            int upleft_corner = grow_it + gcol_it *n_rows;
            //mexPrintf("%i ", upleft_corner);
            float * raw_patch;
            raw_patch = new float[patch_dim]; // NOT initialized to 0s

            // normally not needed
            float * no_mean_norm_patch;
            no_mean_norm_patch = new float[patch_dim]; // NOT initialized to 0s

            float * psyco_patch;
            psyco_patch = new float [patch_dim]; // NOT initialized to 0s

            bool * transf_element;
            transf_element = new bool [4]();

            float * untransformed_patch;
            untransformed_patch = new float[patch_dim]();

            //Regression output
            float * x;
            x = new float[patch_dim]();
            { // keep dim_it only in this scope
                int dim_it = 0;
                for ( int col_it = 0; col_it < patch_side; ++ col_it)
                {
                    for( int row_it = 0; row_it < patch_side; ++ row_it)
                    {
                        raw_patch[dim_it] = image[upleft_corner+ col_it * n_rows + row_it];
                        ++dim_it;
                    }
                }
            } // end of extract_patch scope... function perhaps?
            this->obtainTransfElement(transf_element, raw_patch);
            this->applyTransfElement(psyco_patch,transf_element, raw_patch);

            // BE AWARE - MEAN is subtracted also from the second patch.
            this->noMeanNormalizePatch(no_mean_norm_patch, psyco_patch);

            int index = this->getHash(no_mean_norm_patch);

            // RERANKING AND REGRESSION
            int bucket_size = bucket_sizes[index];
            int acum_bucket_size = this->acum_sizes[index];

            int dmaxpos = 0;
            float maximum = 0.0f;


            if (bucket_size == 0)
            {
                this->matVecMultiplication(this->global_regressor, psyco_patch, x, patch_dim, patch_dim);
            }
            else if (bucket_size == 1)
            {
                this->matVecMultiplication(&this->regressors[acum_bucket_size * patch_dim * patch_dim], psyco_patch, x, patch_dim, patch_dim);
            }
            else
            {
                for(int i = 0; i < bucket_size; i++)
                {
                    float dot_prod = this->dotProduct(no_mean_norm_patch, &this->anchored_features[acum_bucket_size * patch_dim + i * patch_dim], patch_dim);
                    if(i == 0)
                    {
                        maximum = dot_prod;
                        dmaxpos = i;
                    }
                    else if(dot_prod > maximum)
                    {
                        maximum = dot_prod;
                        dmaxpos = i;
                    }
                }
                int R_pos = acum_bucket_size + dmaxpos;

                this->matVecMultiplication(&this->regressors[R_pos * patch_dim * patch_dim], psyco_patch, x, patch_dim,patch_dim);
            }
            this->revertTransfElement(untransformed_patch, transf_element, x);


            for( int col_it = 0; col_it < patch_side; ++col_it)
            {
                for( int row_it = 0; row_it < patch_side; ++row_it )
                {
                    float pixel_value = untransformed_patch[col_it * patch_side + row_it];
#pragma omp atomic
                    image_acc[upleft_corner + col_it * n_rows + row_it] +=  pixel_value;
#pragma omp atomic
                    ++ n_contributions[upleft_corner + col_it * n_rows + row_it];
                }
            }
            //delete buffers
            delete [] raw_patch;
            delete [] no_mean_norm_patch;
            delete [] psyco_patch;
            delete [] transf_element;
            delete [] untransformed_patch;
            delete [] x;


        }
    } // end for each patch
#pragma omp parallel for
    for (int id = 0 ; id < n_rows*n_cols; ++id )
    {
        upscaled_im[id] =  image_acc[id] / ((float)n_contributions[id] + 1e-6f) + image [id];
    } // for each thread

    delete[] n_contributions;
    delete[] image_acc;

}

void PsycoSuperRes::noMeanNormalizePatch(float * output_patch, float * patch)
// Attention:
// This methods substracts the mean from BOTH arguments, also normalizes the first one.
{
    float mean = 0.0f;
    float squared_sum = 0.0f;
    float norm_inv = 0.0f;

    for(int dim_it = 0; dim_it < patch_dim; ++dim_it)
    {
        mean += patch[dim_it];
    } // for each dimension
    mean = mean / patch_dim;

    for(int dim_it = 0; dim_it < patch_dim; ++dim_it)
    {
        patch[dim_it] = patch[dim_it] - mean;
        squared_sum += patch[dim_it] * patch[dim_it];

    } // for each dimension

    norm_inv = 1.0f / sqrt(squared_sum);

    for(int dim_it = 0; dim_it < patch_dim; ++dim_it)
    {
        output_patch[dim_it] = patch[dim_it]*norm_inv;
    } // for each dimension
}

int PsycoSuperRes::getHash(float *patch)
{

    int index = 0;
    int mpow = 1;

    for (int c_it = 0; c_it < n_spheres ; ++c_it)
    {
        float dif = 0.0f;
        for(int dim_it = 0; dim_it < patch_dim; ++dim_it)
        {
            float pc = patch[dim_it] - pivots[dim_it + patch_dim * c_it];
            dif += ( pc * pc) ;
        }
        if( dif <= (this->thresholds[c_it]) )
        {
            index += mpow;
        }
        mpow <<= 1;
    }
    return index;
}

float PsycoSuperRes::dotProduct(const float *patch1, const float *patch2, const int dim)
{
    float dot_product = 0;
    for( int dim_it = 0; dim_it < dim; ++dim_it )
    {

        dot_product += patch1[dim_it] * patch2[dim_it];
    }
    if(dot_product < 0)
    {
        dot_product = -1.0f*dot_product;
    }
    return dot_product;

}

