#include "mex.h"
#include "class_handle.hpp"
#include "Psyco.h"

// The class that we are interfacing to

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{	
    // Get the command string
    char cmd[64];
	if (nrhs < 1 || mxGetString(prhs[0], cmd, sizeof(cmd)))
		mexErrMsgTxt("First input should be a command string less than 64 characters long.");
        
    // New
    if (!strcmp("new", cmd)) {
        // Check parameters
        if (nlhs != 1)
            mexErrMsgTxt("New: One output expected.");
        //assign other inputs
        const int patch_side = (int) mxGetScalar(prhs[1]);
        float * bases = (float *) mxGetData(prhs[2]);

        // Return a handle to a new C++ instance
        plhs[0] = convertPtr2Mat<Psyco>(new Psyco(patch_side, bases));
        return;
    }
    
    // Check there is a second input, which should be the class instance handle
    if (nrhs < 2)
		mexErrMsgTxt("Second input should be a class instance handle.");
    
    // Delete
    if (!strcmp("delete", cmd)) {
        // Destroy the C++ object
        destroyObject<Psyco>(prhs[1]);
        // Warn if other commands were ignored
        if (nlhs != 0 || nrhs != 2)
            mexWarnMsgTxt("Delete: Unexpected arguments ignored.");
        return;
    }
    
    // Get the class instance pointer from the second input
    Psyco *Psyco_instance = convertMat2Ptr<Psyco>(prhs[1]);
    
    // Call the various class methods
    // getDim
    if (!strcmp("getDim", cmd)) {
        // Check parameters
        if (nlhs < 0 || nrhs < 2)
            mexErrMsgTxt("Train: Unexpected arguments.");
        // Call the method
        //int * dim = Psyco_instance->getDim();

        plhs[0] = mxCreateNumericMatrix(1, 1, mxINT32_CLASS, mxREAL);
        int * dim = (int *) mxGetData(plhs[0]);
        *dim = Psyco_instance->getDim();
        return;
    }
    // obtainTransfElement
    if (!strcmp("obtainTransfElement", cmd)) {
        // Check parameters
        if (nlhs < 0 || nrhs < 2)
            mexErrMsgTxt("Train: Unexpected arguments.");
        // Call the method

        //read parameters

        float * patch = (float *) mxGetData(prhs[2]);

        //write paramenters
        plhs[0] = mxCreateLogicalMatrix(4, 1);
        bool * transf_element = (bool *) mxGetData(plhs[0]);
        Psyco_instance->obtainTransfElement(transf_element, patch);
        return;
    }
    // applyTransfElement
    if (!strcmp("applyTransfElement", cmd)) {
        // Check parameters
        if (nlhs < 0 || nrhs < 2)
            mexErrMsgTxt("Train: Unexpected arguments.");
        // Call the method

        //read parameters
        bool *  transf_element = (bool *) mxGetData(prhs[2]);
        float * patch = (float *) mxGetData(prhs[3]);

        //write paramenters
        plhs[0] = mxCreateNumericMatrix(Psyco_instance->getDim(), 1, mxSINGLE_CLASS, mxREAL);
        float * transf_patch = (float *) mxGetData(plhs[0]);
        Psyco_instance->applyTransfElement(transf_patch, transf_element, patch);
        return;
    }

    // revertTransfElement
    if (!strcmp("revertTransfElement", cmd)) {
        // Check parameters
        if (nlhs < 0 || nrhs < 2)
            mexErrMsgTxt("Train: Unexpected arguments.");
        // Call the method

        //read parameters
        bool *  transf_element = (bool *) mxGetData(prhs[2]);
        float * transf_patch = (float *) mxGetData(prhs[3]);

        //write paramenters
        plhs[0] = mxCreateNumericMatrix(Psyco_instance->getDim(), 1, mxSINGLE_CLASS, mxREAL);
        float * reverted_patch = (float *) mxGetData(plhs[0]);
        Psyco_instance->revertTransfElement(reverted_patch, transf_element, transf_patch);
        return;
    }

    // applyTransfPar
    if (!strcmp("applyTransfPar", cmd)) {
        // Check parameters
        if (nlhs < 0 || nrhs < 2)
            mexErrMsgTxt("Train: Unexpected arguments.");
        // Call the method

        //read parameters
        float * patches = (float *) mxGetData(prhs[2]);
        const int n_patches = mxGetN(prhs[2]);

        //write paramenters
        plhs[0] = mxCreateNumericMatrix(Psyco_instance->getDim(), n_patches, mxSINGLE_CLASS, mxREAL);
        plhs[1] = mxCreateLogicalMatrix(4, n_patches);

        float * transformed_patches = (float *) mxGetData(plhs[0]);
        bool * transf_elements = (bool *) mxGetData(plhs[1]);

        Psyco_instance->applyTransfPar(transformed_patches, transf_elements, patches,n_patches);
        return;
    }


    // Test    
    if (!strcmp("getSide", cmd)) {
        // Check parameters
        if (nlhs < 0 || nrhs < 2)
            mexErrMsgTxt("Test: Unexpected arguments.");
        // Call the method
        Psyco_instance->getSide();
        return;
    }
    
    // Got here, so command not recognized
    mexErrMsgTxt("Command not recognized.");
}
